#!/usr/bin/env python
# Rayleigh-Taylor Instability
# Scripts arguments:
# 1. Flow resolution
# 2. Scalar resolution
# 3. Dictionary for devices id: (mpi rank: device id)
# 4. Is data output
# 5. Low density
# 6. High density
# mpirun -np 10 python ./RTI_new.py "[129,129,129]" "[257,257,257]" "{0:0,5:1}" "True" "1" "3"
import sys
USER_NB_ELEM_UW = eval(sys.argv[1])
USER_NB_ELEM_S = eval(sys.argv[2])
USER_RANK_DEVICE_ID = eval(sys.argv[3])
IS_OUTPUT = eval(sys.argv[4])
LOW_DENSITY = eval(sys.argv[5])
HIGH_DENSITY = eval(sys.argv[6])
import hysop
# import hysop.gpu
# hysop.gpu.CL_PROFILE = True
from hysop.constants import np, HDF5, ASCII, HYSOP_MPI_REAL
from hysop.mpi.main_var import MPI, main_size, main_rank, main_comm
from hysop.tools.parameters import MPIParams, Discretization, IOParams
from hysop.problem.simulation import Simulation
from hysop.fields.variable_parameter import VariableParameter
from hysop.methods_keys import TimeIntegrator, Interpolation, Remesh,\
    Support, Splitting, MultiScale, MultiScale, SpaceDiscretisation, \
    GhostUpdate, Scales, dtCrit, ExtraArgs
from hysop.numerics.integrators.runge_kutta2 import RK2 as RK
from hysop.numerics.integrators.runge_kutta3 import RK3 as RK3
from hysop.numerics.finite_differences import FD_C_4
from hysop.numerics.remeshing import L6_4 as remesh_formula
from hysop.numerics.remeshing import L2_1
from hysop.numerics.remeshing import Linear
from hysop.operator.advection import Advection
from hysop.operator.diffusion import Diffusion
from hysop.operator.stretching import Stretching
from hysop.operator.poisson import Poisson
from hysop.operator.baroclinic_from_rhs import BaroclinicFromRHS
from hysop.operator.differential import Curl
from hysop.operator.multiresolution_filter import MultiresolutionFilter
from hysop.operator.multiphase_gradp import MultiphaseGradP
from hysop.operator.multiphase_baroclinic_rhs import MultiphaseBaroclinicRHS
from hysop.operator.adapt_timestep import AdaptTimeStep
from hysop.operator.custom import CustomMonitor
from hysop.operator.redistribute_inter import RedistributeInter
from hysop.operator.redistribute_intra import RedistributeIntra
from hysop.gpu.gpu_transfer import DataTransfer
from hysop.domain.subsets import SubBox
from hysop.operator.hdf_io import HDF_Writer
from hysop.operator.energy_enstrophy import EnergyEnstrophy
import hysop.tools.numpywrappers as npw
from hysop.tools.io_utils import IO
from hysop.operator.spectrum import Spectrum
pi = np.pi
cos = np.cos
sin = np.sin
exp = np.exp
abs = np.abs
tanh = np.tanh

TASK_UW = 1
TASK_SCALAR = 2
PROC_TASKS = [TASK_UW, ] * main_size
for p in USER_RANK_DEVICE_ID:
    PROC_TASKS[p] = TASK_SCALAR
try:
    DEVICE_ID = USER_RANK_DEVICE_ID[main_rank]
except KeyError:
    DEVICE_ID = None
print USER_RANK_DEVICE_ID
out_freq = 10
# Physical parameters:
# Flow viscosity
VISCOSITY = 5e-5
# Schmidt number
SC = ((1. * USER_NB_ELEM_S[0] - 1.)/(1. * USER_NB_ELEM_UW[0] - 1.))**2
# Scalar diffusivity
DIFF_COEFF_SCAL = VISCOSITY / SC
MAX_DT = 0.05  # ok with 0.05
MAX_CFL = 1.5  # ok with 1.5 (lcfl bound)
MAX_LCFL = 0.15 # ok with 0.2
#MAX_CFL_VEC = [1.5, 0.6, 0.5]


width = 0.01
ampl3 = 0.3
ampl = 0.05
ampl2 = 0.05

# NS
# CPU ::  Velo(C_2d), Vorti(C_2d) > Advec > Vorti(C_2d)
# CPU ::  Velo(C_2d,2G), Vorti(C_2d,2G) > Stretching > Vorti(C_2d,2G)
# CPU ::  Vorti(C_2d,2G) > Penalisation > Vorti(C_2d,2G)
# CPU ::  Vorti(C_1d) > Diffusion > Vorti(C_1d)
# CPU ::  Vorti(C_1d) > Poisson > Velo(C_1d)
# CPU ::  Velo(C_2d,2G), Vorti(C_2d,2G) > Dt
# Scal
# GPU ::  Velo(C_2d,1G), Scal(F_2d) > Advec > Scal(F_2d)
# Baroclinic
# GPU ::  Velo(C_2d,2G) > GradP > gradp(C_2d,2G)
# GPU ::  Rho(F_2d), gradp(C_2d,2G) > BaroclinicRHS > rhs(F_2d)
# GPU ::  rhs(F_2d) > Filter > rhs(C_2d,2G)
# CPU ::  rhs(C_2d,2G), Vorti(C_2d,2G) > Baroclinic > Vorti(C_2d,2G)

ctime = MPI.Wtime()

# Domain
box = hysop.Box(length=[1., 1., 1.], origin=[0., 0., 0.],
                proc_tasks=PROC_TASKS)
mpi_params = MPIParams(comm=box.comm_task, task_id=PROC_TASKS[main_rank])
mpi_params_S = MPIParams(comm=box.comm_task, task_id=TASK_SCALAR)
mpi_params_UW = MPIParams(comm=box.comm_task, task_id=TASK_UW)

SIGMA=1e-3
NOISE=1e-4

def computeVelo(res, x, y, z, t):
    zz = exp(-(z-0.5) * (z-0.5) / SIGMA) * NOISE
    from create_random_arrays import random_init
    randX, randY, randZ = random_init(res[0].shape, box.comm_task)
    res[0][...] = zz * randX
    res[1][...] = zz * randY
    res[2][...] = zz * randZ
    return res

def initScal(res, x, y, z, t):
    res[0][...] = z
    return res


temp_maxvelo = npw.zeros((3, ))
maxvelo_values = npw.zeros((3, ))


def calc_maxvelo(simu, v):
    temp_maxvelo[0] = np.max(np.abs(v[0].data[0]))
    temp_maxvelo[1] = np.max(np.abs(v[0].data[1]))
    temp_maxvelo[2] = np.max(np.abs(v[0].data[2]))
    v[0].topology.comm.Allreduce(sendbuf=[temp_maxvelo, 3, HYSOP_MPI_REAL],
                                 recvbuf=[maxvelo_values, 3, HYSOP_MPI_REAL],
                                 op=MPI.MAX)
    return maxvelo_values


# Fields
velo = hysop.Field(domain=box, formula=computeVelo,
                   name='Velocity', is_vector=True)
vorti = hysop.Field(domain=box,
                    name='Vorticity', is_vector=True)
scal = hysop.Field(domain=box, formula=initScal,
                   name='Scalar', is_vector=False)
gradp = hysop.Field(domain=box,
                    name='GradP', is_vector=True)
baroclinic_rhs = hysop.Field(domain=box,
                             name='B_rhs', is_vector=True)

data = {'dt': 0.001}
dt = VariableParameter(data)
simu = Simulation(tinit=0.0, tend=2.5, timeStep=0.001, iterMax=10000)

# Flow discretizations:
d_C = Discretization(USER_NB_ELEM_UW)
d_C_2G = Discretization(USER_NB_ELEM_UW, [2, ] * 3)
# Scalar discretization
d_F = Discretization(USER_NB_ELEM_S)

# Topologies
topo_C_2d = None
topo_C_2d_2G = None
topo_C_1d = None
topo_F_2d = None
topo_C_2d_2G = box.create_topology(
    d_C_2G, dim=1, mpi_params=mpi_params)
topo_C_2d = box.create_topology(
    d_C, dim=2, mpi_params=mpi_params)
topo_C_1d = box.create_topology(
    d_C, dim=1, mpi_params=mpi_params)
topo_F_2d = box.create_topology(
    d_F, dim=1, mpi_params=mpi_params)
###### WARNING This topo is 1D to be fftw compliant.


# Operators
# GPU operators
advec_scal_method = {TimeIntegrator: RK,
                     Interpolation: Linear,
                     Remesh: remesh_formula,
                     Support: 'gpu_1k',
                     Splitting: 'o2',
                     MultiScale: Linear}
if PROC_TASKS.count(TASK_SCALAR) == 1:
    advec_scal_method[ExtraArgs] = {'device_id': DEVICE_ID,
                                    'device_type': 'gpu'}
else:
    advec_scal_method[ExtraArgs] = {'max_cfl': MAX_CFL,
                                    'max_velocity': [2.5, 2.5, 3.5], # for rho=1<->3
                                    'max_dt': MAX_DT,
                                    'device_id': DEVICE_ID,
                                    'device_type': 'gpu'}
advec_scal = Advection(velo,
                       discretization=topo_C_2d_2G,
                       variables={scal: topo_F_2d},
                       mpi_params=mpi_params_S,
                       method=advec_scal_method)
diffusion_scal = Diffusion(viscosity=DIFF_COEFF_SCAL,
                           vorticity=scal,
                           discretization=topo_F_2d,
                           mpi_params=mpi_params_S,
                           method={Support: 'gpu',
                                   SpaceDiscretisation: 'fd',
                                   ExtraArgs: {'device_id': DEVICE_ID,
                                               'device_type': 'gpu'}})
diffusion_scal.name += '_(Scalar)'
# tanh(x) is in [-1:1] and -1 stand for the flow.
# We want [LOW_DENSITY:HIGH_DENSITY]
scal_to_rho = '{0:f}*({1}*tanh(100.0*x-50.0)+0.5)+{2:f}'.format(
    abs(LOW_DENSITY - HIGH_DENSITY),
    0.5 if LOW_DENSITY < HIGH_DENSITY else -0.5,
    min(LOW_DENSITY, HIGH_DENSITY))
baroclinic_rhs_op = MultiphaseBaroclinicRHS(
    baroclinic_rhs, scal, gradp,
    variables={baroclinic_rhs: topo_F_2d,
               gradp: topo_C_2d_2G,
               scal: topo_F_2d},
    method={Support: 'gpu',
            SpaceDiscretisation: FD_C_4,
            ExtraArgs: {'density_func': scal_to_rho,
                        'device_id': DEVICE_ID,
                        'device_type': 'gpu'}},
    mpi_params=mpi_params_S)
filter_scal = MultiresolutionFilter(
    d_in=topo_F_2d, d_out=topo_C_2d_2G,
    variables={baroclinic_rhs: topo_C_2d_2G},
    method={Remesh: L2_1,
            Support: 'gpu',
            ExtraArgs: {'device_id': DEVICE_ID,
                        'device_type': 'gpu'}},
    mpi_params=mpi_params_S)
gradp_op = MultiphaseGradP(velocity=velo, gradp=gradp, viscosity=VISCOSITY,
                           discretization=topo_C_2d_2G,
                           mpi_params=mpi_params_S,
                           method={SpaceDiscretisation: FD_C_4,
                                   ExtraArgs: {'gravity': [0., 0., -1.]}})

# CPU operators
advec = Advection(velo,
                  discretization=topo_C_2d,
                  variables={vorti: topo_C_2d},
                  mpi_params=mpi_params_UW,
                  method={Scales: 'p_64', MultiScale: 'L4_4'})
stretch = Stretching(velo, vorti, discretization=topo_C_2d_2G,
                     mpi_params=mpi_params_UW)
baroclinic = BaroclinicFromRHS(vorti, baroclinic_rhs,
                               discretization=topo_C_2d_2G,
                               mpi_params=mpi_params_UW)
diffusion = Diffusion(variables={vorti: topo_C_1d},
                      viscosity=VISCOSITY,
                      mpi_params=mpi_params_UW)
poisson = Poisson(velo, vorti, discretization=topo_C_1d,
                  mpi_params=mpi_params_UW)
c = Curl(velo, vorti, discretization=topo_C_1d,
         method={SpaceDiscretisation: 'fftw', GhostUpdate: True},
         mpi_params=mpi_params_UW)
#dt_output = None
#if IS_OUTPUT:
dt_output = IOParams(frequency=1, filename='dt.dat', fileformat=ASCII)
dt_adapt = AdaptTimeStep(velo, vorti,
                         simulation=simu,
                         time_range=[10, np.infty],
                         discretization=topo_C_2d_2G,
                         method={TimeIntegrator: RK3,
                                 SpaceDiscretisation: FD_C_4,
                                 dtCrit: ['gradU', 'cfl']},
                         lcfl=MAX_LCFL, maxdt=MAX_DT,
                         cfl=MAX_CFL,
                         io_params=dt_output,
                         mpi_params=mpi_params_UW)

# Operators discretizations
if box.is_on_task(TASK_SCALAR):
    for op in (advec_scal, gradp_op, filter_scal,
               baroclinic_rhs_op, diffusion_scal):
        op.discretize()
if box.is_on_task(TASK_UW):
    for op in (advec, stretch, diffusion, poisson, c, dt_adapt, baroclinic):
        op.discretize()

if IS_OUTPUT:
    # Defining subdomains
    L_Vx = 1. - 1. / (1. * USER_NB_ELEM_UW[0])
    L_Sx = 1. - 1. / (1. * USER_NB_ELEM_S[0])
    L_Vy = 1. - 1. / (1. * USER_NB_ELEM_UW[1])
    L_Sy = 1. - 1. / (1. * USER_NB_ELEM_S[1])
    L_Vz = 1. - 1. / (1. * USER_NB_ELEM_UW[2])
    L_Sz = 1. - 1. / (1. * USER_NB_ELEM_S[2])
    XY_plane_v = SubBox(origin=[0., 0., 1.], length=[L_Vx, L_Vy, 0.],
                        parent=box)
    XZ_plane_v = SubBox(origin=[0., 0.5, 0.], length=[L_Vx, 0., L_Vz],
                        parent=box)
    YZ_plane_s = SubBox(origin=[0.5, 0., 0.], length=[0., L_Sy, L_Sz],
                        parent=box)
    XZ_plane_s = SubBox(origin=[0., 0.5, 0.], length=[L_Sx, 0., L_Sz],
                        parent=box)
    # Defining output operators
    p_velo = HDF_Writer(variables={velo: topo_C_1d, vorti: topo_C_1d},
                        mpi_params=mpi_params_UW,
                        io_params=IOParams(frequency=out_freq,
                                           filename='flow',
                                           fileformat=HDF5))
    p_scal_yz = HDF_Writer(variables={scal: topo_F_2d},
                           var_names={scal: 'Scalar'},
                           subset=YZ_plane_s,
                           mpi_params=mpi_params_S,
                           io_params=IOParams(frequency=out_freq,
                                              filename='scal_YZ',
                                              fileformat=HDF5))
    p_scal_xz = HDF_Writer(variables={scal: topo_F_2d},
                           var_names={scal: 'Scalar'},
                           subset=XZ_plane_s,
                           mpi_params=mpi_params_S,
                           io_params=IOParams(frequency=out_freq,
                                              filename='scal_XZ',
                                              fileformat=HDF5))
    p_scal = HDF_Writer(variables={scal: topo_F_2d},
                        var_names={scal: 'Scalar'},
                        mpi_params=mpi_params_S,
                        io_params=IOParams(frequency=out_freq,
                                           filename='scal',
                                           fileformat=HDF5))
    p_scal_yz.name += '_(Scalar)'
    p_scal_xz.name += '_(Scalar)'
    p_scal.name += '_(Scalar)'
    energy = EnergyEnstrophy(velocity=velo,
                             vorticity=vorti,
                             discretization=topo_C_1d,
                             mpi_params=mpi_params_UW,
                             io_params=IOParams(frequency=1,
                                                filename='energy.dat',
                                                fileformat=ASCII))
    maxvelo = CustomMonitor(variables={velo: topo_C_1d},
                            function=calc_maxvelo,
                            res_shape=3,
                            mpi_params=mpi_params_UW,
                            io_params=IOParams(frequency=1,
                                               filename='maxvelo.dat',
                                               fileformat=ASCII))
    if box.is_on_task(TASK_UW):
        for op in (p_velo, energy, maxvelo):
            op.discretize()
    if box.is_on_task(TASK_SCALAR):
        for op in (p_scal_yz, p_scal_xz, p_scal):
            op.discretize()

# Redistribute operators
# CPU redistributes
W_C_2d_to_2G = RedistributeIntra(variables=[vorti],
                                 source=advec, target=stretch,
                                 mpi_params=mpi_params_UW)
W_C_2d_to_2G.name += '_W_C_2d_to_2G'
W_C_2d_2G_to_1d = RedistributeIntra(variables=[vorti],
                                    source=baroclinic,  # stretch,
                                    target=diffusion,
                                    mpi_params=mpi_params_UW)
W_C_2d_2G_to_1d.name += '_W_C_2d_2G_to_1d'
W_C_1d_to_C_2d_2G = RedistributeIntra(variables=[vorti],
                                      source=diffusion, target=dt_adapt,
                                      mpi_params=mpi_params_UW)
W_C_1d_to_C_2d_2G.name += '_W_C_1d_to_C_2d_2G'
W_C_1d_to_C_2d = RedistributeIntra(variables=[vorti],
                                   source=diffusion, target=advec,
                                   mpi_params=mpi_params_UW)
W_C_1d_to_C_2d.name += "_W_C_1d_to_C_2d"
U_C_1d_to_C_2d_2G = RedistributeIntra(variables=[velo],
                                      source=poisson, target=stretch,
                                      run_till=[stretch, dt_adapt],
                                      mpi_params=mpi_params_UW)
U_C_1d_to_C_2d_2G.name += '_U_C_1d_to_C_2d_2G'
U_C_1d_to_C_2d = RedistributeIntra(variables=[velo],
                                   source=poisson, target=advec,
                                   mpi_params=mpi_params_UW)
U_C_1d_to_C_2d.name += '_U_C_1d_to_C_2d'
toDevU = DataTransfer(source=topo_C_2d_2G, target=advec_scal,
                      variables={velo: topo_C_2d_2G},
                      run_till=[advec_scal, gradp_op],
                      mpi_params=mpi_params_S)
toDevU.name += '_ToDev_U'
U_CPU_to_GPU = RedistributeInter(variables=[velo],
                                 parent=main_comm,
                                 source=topo_C_1d, target=topo_C_2d_2G,
                                 source_id=TASK_UW, target_id=TASK_SCALAR,
                                 run_till=[toDevU])
rhs_GPU_to_CPU = RedistributeInter(variables=[baroclinic_rhs],
                                   parent=main_comm,
                                   source=topo_C_2d_2G, target=topo_C_2d_2G,
                                   source_id=TASK_SCALAR, target_id=TASK_UW,
                                   run_till=[baroclinic])
rhs_GPU_to_CPU.name += '_rhs_GPU_to_CPU'  # ok
toHostRHS = DataTransfer(source=topo_C_2d_2G,
                         target=rhs_GPU_to_CPU,
                         variables={baroclinic_rhs: topo_C_2d_2G},
                         mpi_params=mpi_params_S,)
toHostRHS.name += '_ToHost_RHS'
toDevGradP = DataTransfer(source=topo_C_2d_2G, target=baroclinic_rhs_op,
                          variables={gradp: topo_C_2d_2G},
                          mpi_params=mpi_params_S)
toDevGradP.name += '_ToDev_GradP'

if IS_OUTPUT:
    toHostS = DataTransfer(source=topo_F_2d, target=p_scal_yz,
                           variables={scal: topo_F_2d},
                           mpi_params=mpi_params_S,
                           freq=out_freq,
                           run_till=[p_scal_xz, p_scal_yz, p_scal])
    toHostS.name += '_ToHost_S'

# Operators setup
if box.is_on_task(TASK_SCALAR):
    for op in (advec_scal, gradp_op,
               baroclinic_rhs_op, filter_scal, diffusion_scal):
        op.setup()
    if IS_OUTPUT:
        for op in (p_scal_yz, p_scal_xz, p_scal, toHostS):
            op.setup()
    for op in (toDevU, toHostRHS, toDevGradP):
        op.setup()
if box.is_on_task(TASK_UW):
    for op in (advec, stretch, diffusion, poisson, c, dt_adapt, baroclinic):
        op.setup()
    if IS_OUTPUT:
        for op in (p_velo, energy, maxvelo):
            op.setup()
    for op in (W_C_2d_to_2G, W_C_2d_2G_to_1d,
               W_C_1d_to_C_2d, W_C_1d_to_C_2d_2G,
               U_C_1d_to_C_2d, U_C_1d_to_C_2d_2G):
        op.setup()
# Wait for all operators setup before setup the intra-comm redistribute
main_comm.Barrier()
U_CPU_to_GPU.setup()
rhs_GPU_to_CPU.setup()

# Operators list
if IS_OUTPUT:
    operators_list = [toDevU, gradp_op, toDevGradP,
                      advec_scal, diffusion_scal,
                      toHostS, p_scal_yz, p_scal_xz, p_scal,
                      baroclinic_rhs_op, filter_scal,
                      toHostRHS, rhs_GPU_to_CPU,
                      advec, W_C_2d_to_2G, stretch, baroclinic,
                      W_C_2d_2G_to_1d, diffusion, poisson,
                      p_velo, energy, maxvelo,
                      U_CPU_to_GPU,
                      U_C_1d_to_C_2d, U_C_1d_to_C_2d_2G,
                      W_C_1d_to_C_2d, W_C_1d_to_C_2d_2G, dt_adapt]
else:
    operators_list = [toDevU, advec_scal, diffusion_scal,
                      advec, W_C_2d_to_2G, stretch,
                      W_C_2d_2G_to_1d, diffusion, poisson,
                      energy, maxvelo,
                      U_CPU_to_GPU,
                      U_C_1d_to_C_2d, U_C_1d_to_C_2d_2G,
                      W_C_1d_to_C_2d, W_C_1d_to_C_2d_2G, dt_adapt]

# Fields initializations
if box.is_on_task(TASK_SCALAR):
    scal.initialize(topo=topo_F_2d)
    advec_dirX = advec_scal.advec_dir[0].discrete_op
    gpu_scal = advec_dirX.fields_on_grid[0]
    gpu_pscal = advec_dirX.fields_on_part[gpu_scal][0]
    diffusion_scal.discrete_op.set_field_tmp(gpu_pscal)
    if IS_OUTPUT:
        p_scal_yz.apply(simu)
        p_scal_xz.apply(simu)
        p_scal.apply(simu)
if box.is_on_task(TASK_UW):
    velo.initialize(topo=topo_C_1d)
    c.apply(simu)
    poisson.apply(simu)
    U_C_1d_to_C_2d.apply(simu)
    U_C_1d_to_C_2d_2G.apply(simu)
    W_C_1d_to_C_2d.apply(simu)
    W_C_1d_to_C_2d_2G.apply(simu)
    if IS_OUTPUT:
        p_velo.apply(simu)
        # p_velo_xy.apply(simu)
        # p_velo_xz.apply(simu)
U_CPU_to_GPU.apply(simu)
U_CPU_to_GPU.wait()
if box.is_on_task(TASK_SCALAR):
    toDevU.apply(simu)
    toDevU.wait()
    gradp_op.initialize_velocity()

simu.initialize()
setup_time = MPI.Wtime() - ctime
main_comm.Barrier()

# Solve
solve_time = 0.
while not simu.isOver:
    ctime = MPI.Wtime()
    if main_rank == 0:
        simu.printState()
    for op in operators_list:
        if box.is_on_task(op.task_id()):
            op.apply(simu)
    if box.is_on_task(TASK_SCALAR):
        # Wait gpu operations on scalar
        advec_scal.advec_dir[0].discreteFields[scal].wait()
    solve_time += MPI.Wtime() - ctime
    dt_adapt.wait()
    # Synchronize threads
    main_comm.Barrier()
    simu.advance()

U_CPU_to_GPU.wait()
main_comm.Barrier()
nb_ite = simu.currentIteration
simu.finalize()


if IS_OUTPUT:
    if box.is_on_task(TASK_SCALAR):
        p_scal_yz.apply(simu)
        p_scal_xz.apply(simu)
        p_scal.apply(simu)
    if box.is_on_task(TASK_UW):
        p_velo.apply(simu)
        # p_velo_xy.apply(simu)
        # p_velo_xz.apply(simu)

for op in operators_list:
    if box.is_on_task(op.task_id()):
        op.finalize()

velo.finalize()
if box.is_on_task(TASK_SCALAR):
    scal.finalize()
if box.is_on_task(TASK_UW):
    vorti.finalize()
main_comm.Barrier()
